from ..generic.general_methods import aedt_exception_handler, retry_ntimes
from .PrimitivesCircuit import CircuitComponents


class NexximComponents(CircuitComponents):
    """NexximComponents class.
    
    This class is for managing all circuit components for Nexxim.
    """
    @property
    def design_libray(self):
        return "Nexxim Circuit Elements"

    @property
    def tab_name(self):
        return "PassedParameterTab"

    @aedt_exception_handler
    def __getitem__(self, partname):
        """Get the object ID if partname is an integer or object name if a string.
        Parameters
        ----------
        partname: int or str
            Part ID or object name. 
        
        Returns
        -------
        type
            Part object details.
        """
        if type(partname) is int:
            return self.components[partname]
        for el in self.components:
            if self.components[el].name == partname or self.components[el].composed_name == partname or el == partname:
                return self.components[el]

        return None

    def __init__(self, parent, modeler):
        CircuitComponents.__init__(self, parent, modeler)
        self._parent = parent
        self.modeler = modeler
        self._currentId = 0
        pass

    @aedt_exception_handler
    def create_3dlayout_subcircuit(self, sourcename):
        """Create a new subcircuit.

        Parameters
        ----------
        sourcename : str
            Name of the source design.

        Returns
        -------
        type
            el, composed_name if succeeded or False

        """
        self._parent._oproject.CopyDesign(sourcename)
        self.oeditor.PasteDesign(0, ["NAME:Attributes", "Page:=", 1, "X:=", 0, "Y:=", 0, "Angle:=", 0, "Flip:=", False])
        self.refresh_all_ids()
        for el in self.components:
            if sourcename in self.components[el].composed_name:
                return el, self.components[el].composed_name
        return False

    @aedt_exception_handler
    def create_field_model(self, design_name, solution_name, pin_names, model_type="hfss", posx=0, posy=1):
        """Create a field model.

        Parameters
        ----------
        design_name : str
            Name of the design.
        solution_name: str
            Name  of the solution.
        pin_names : list
            List of the pins.
        model_type: str, optional
            Type of the model. The default is ``"hfss"``.
        posx : float, optional
            Position on the X axis. The default is ``0``.    
        posy : float, optional.
            Position on the Y axis. The default is ``1``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        id = self.create_unique_id()
        component_name = design_name + "_" + str(id)
        arg = ["NAME: " +component_name ,"Name:="	, component_name,
               "ModTime:="		, 1589868932,
               "Library:="		, "",
               "LibLocation:="		, "Project",
               "ModelType:="		, model_type,
               "Description:="		, "",
               "ImageFile:="		, "",
               "SymbolPinConfiguration:=", 0,
               [
                   "NAME:PortInfoBlk"
               ],
               [
                   "NAME:PortOrderBlk"
               ],
               "DesignName:="		, design_name,
               "SolutionName:="	, solution_name,
               "NewToOldMap:="		, [],
               "OldToNewMap:="		, [],
               "PinNames:="		, pin_names,
               [
                   "NAME:DesignerCustomization",
                   "DCOption:="		, 0,
                   "InterpOption:="	, 0,
                   "ExtrapOption:="	, 1,
                   "Convolution:="		, 0,
                   "Passivity:="		, 0,
                   "Reciprocal:="		, False,
                   "ModelOption:="		, "",
                   "DataType:="		, 1
               ],
               [
                   "NAME:NexximCustomization",
                   "DCOption:="		, 3,
                   "InterpOption:="	, 1,
                   "ExtrapOption:="	, 3,
                   "Convolution:="		, 0,
                   "Passivity:="		, 0,
                   "Reciprocal:="		, False,
                   "ModelOption:="		, "",
                   "DataType:="		, 2
               ],
               [
                   "NAME:HSpiceCustomization",
                   "DCOption:="		, 1,
                   "InterpOption:="	, 2,
                   "ExtrapOption:="	, 3,
                   "Convolution:="		, 0,
                   "Passivity:="		, 0,
                   "Reciprocal:="		, False,
                   "ModelOption:="		, "",
                   "DataType:="		, 3
               ],
               "NoiseModelOption:="	, "External",
               "WB_SystemID:="		, design_name,
               "IsWBModel:="		, False,
               "filename:="		, "",
               "numberofports:="	, len(pin_names),
               "Simulate:="		, False,
               "CloseProject:="	, False,
               "SaveProject:="		, True,
               "InterpY:="		, True,
               "InterpAlg:="		, "auto",
               "IgnoreDepVars:="	, False,
               "Renormalize:="		, False,
               "RenormImpedance:="	, 50
               ]
        self.o_model_manager.Add(arg)
        arg = [
            "NAME:" + component_name,
            "Info:=",
            ["Type:=", 8, "NumTerminals:=", len(pin_names), "DataSource:=", "", "ModifiedOn:=", 1589868933,
             "Manufacturer:=", "",
             "Symbol:=", "", "ModelNames:=", "", "Footprint:=", "", "Description:=", "", "InfoTopic:=", "",
             "InfoHelpFile:=", "", "IconFile:=", "", "Library:=", "", "OriginalLocation:=", "Project",
             "IEEE:=", "", "Author:=", "", "OriginalAuthor:=", "", "CreationDate:=", 1589868933, "ExampleFile:=",
             "", "HiddenComponent:=", 0, "CircuitEnv:=", 0, "GroupID:=", 0],
            "CircuitEnv:="	, 0,
            "Refbase:="		, "S",
            "NumParts:="		, 1,
            "ModSinceLib:="		, False]
        id = 2
        for pn in pin_names:
            arg.append("Terminal:=")
            arg.append([pn ,pn, "A", False, id, 1, "", "Electrical", "0"])
            id += 1
        arg.append(["NAME:Properties", "TextProp:=", ["Owner", "RD", "", model_type.upper()]])
        arg.append("CompExtID:="), arg.append(5)
        arg.append(["NAME:Parameters", "TextProp:="	, ["ModelName" ,"RD", "", "FieldSolver"],
                    "MenuProp:=", ["CoSimulator", "SD", "", "Default", 0],
                    "ButtonProp:=", ["CosimDefinition", "SD", "", "Edit", "Edit", 40501, "ButtonPropClientData:=", []]])
        arg.append(["NAME:CosimDefinitions",
                    ["NAME:CosimDefinition", "CosimulatorType:=", 103, "CosimDefName:=", "Default", "IsDefinition:=",
                     True, "Connect:=", True,
                     "ModelDefinitionName:=", component_name, "ShowRefPin2:=", 2, "LenPropName:=", ""],
                    "DefaultCosim:=", "Default"])

        self.o_component_manager.Add(arg)
        self._parent.odesign.AddCompInstance(component_name)
        self.refresh_all_ids()
        for el in self.components:
            if component_name in self.components[el].composed_name:
                return el, self.components[el].composed_name
        return False

    @aedt_exception_handler
    def create_resistor(self, compname=None, value=50, xpos=0, ypos=0,angle=0, use_instance_id_netlist=False):
        """Create a new resistor.

        Parameters
        ----------
        compname : str, optional
            Name of the resistor. The default is ``None``.
        value : float, optional
            Resistance in ohms. The default is ``50``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.

        Returns
        -------
        type
            ID of the resistor.
        str
            Name of the resistor.
        """
        cmpid, cmpname = self.create_component(compname, xpos=xpos, ypos=ypos, angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("R", value)
        return cmpid, cmpname

    @aedt_exception_handler
    def create_inductor(self, compname=None,value=50, xpos=0, ypos=0,angle=0, use_instance_id_netlist=False):
        """Create a new inductor.

        Parameters
        ----------
        compname : str, optional
            Name of the inductor. The default is ``None``.
        value : float, optional
            The default is ``50``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the X axis. The default is ``0``.    
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.

        Returns
        -------
        type
            ID of the inductor.
        str
            Name of the inductor.
        """
        cmpid, cmpname = self.create_component(compname, component_library="Inductors", component_name="IND_", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("L", value)

        return cmpid, cmpname

    @aedt_exception_handler
    def create_capacitor(self, compname=None,value=50, xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new capacitor.

        Parameters
        ----------
        compname : str, optional
            Name of the capacitor. The default is ``None``.
        value : float, optional
            The default is ``50``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.
        
        Returns
        -------
        type
            ID of the capacitor.
        str
            Name of the capacitor.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Capacitors", component_name="CAP_", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("C", value)
        return cmpid, cmpname

    @aedt_exception_handler
    def create_voltage_dc(self, compname=None, value=1, xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new voltage DC source.

        Parameters
        ----------
        compname : str, optional
            Name of the voltage DC source. The default is ``None``.
        value : float, optional
            The default is ``50``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.

        Returns
        -------
        type
            ID of the voltage DC source.
        str
            Name of the voltage DC source.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Independent Sources", component_name="V_DC", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("DC", value)
        return cmpid, cmpname

    @aedt_exception_handler
    def create_current_pulse(self, compname=None, value_lists=[], xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new current pulse.

        Parameters
        ----------
        compname : str, optional
            Name of the current pulse. The default is ``None``.
        value_lists : list, optional
            List of values for the current pulse. The default is ``[]``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.
        
        Returns
        -------
        type
            ID of the current pulse.
        str
            Name of the current pulse.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Independent Sources", component_name="I_PULSE", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        if len(value_lists) > 0:
            self.components[cmpid].set_property("I1", value_lists[0])
        if len(value_lists) > 1:
            self.components[cmpid].set_property("I2", value_lists[1])
        if len(value_lists) > 2:
            self.components[cmpid].set_property("TD", value_lists[2])
        if len(value_lists) > 3:
            self.components[cmpid].set_property("TR", value_lists[3])
        if len(value_lists) > 4:
            self.components[cmpid].set_property("TF", value_lists[4])
        if len(value_lists) > 5:
            self.components[cmpid].set_property("PW", value_lists[5])
        if len(value_lists) > 6:
            self.components[cmpid].set_property("PER", value_lists[6])


        return cmpid, cmpname

    @aedt_exception_handler
    def create_voltage_pulse(self, compname=None, value_lists=[], xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new voltage pulse.

        Parameters
        ----------
        compname : str, optional
            Name of the voltage pulse. The default is ``None``.
        value_lists : list, optional
            List of values for the voltage pulse. The default is ``[]``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.
        
        Returns
        -------
        type
            ID of the voltage pulse.
        str
            Name of the voltage pulse.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Independent Sources", component_name="V_PULSE", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        if len(value_lists) > 0:
            self.components[cmpid].set_property("V1", value_lists[0])
        if len(value_lists) > 1:
            self.components[cmpid].set_property("V2", value_lists[1])
        if len(value_lists) > 2:
            self.components[cmpid].set_property("TD", value_lists[2])
        if len(value_lists) > 3:
            self.components[cmpid].set_property("TR", value_lists[3])
        if len(value_lists) > 4:
            self.components[cmpid].set_property("TF", value_lists[4])
        if len(value_lists) > 5:
            self.components[cmpid].set_property("PW", value_lists[5])
        if len(value_lists) > 6:
            self.components[cmpid].set_property("PER", value_lists[6])


        return cmpid, cmpname


    @aedt_exception_handler
    def create_current_dc(self, compname=None, value=1, xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new current DC source.

        Parameters
        ----------
        compname : str, optional
            Name of the current DC source. The default is ``None``.
        value : float, optional
            Value for the current DC source. The default is ``1``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.

        Returns
        -------
        type
            ID of the current DC source.
        str
            Name of the current DC source.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Independent Sources", component_name="I_DC", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("DC", value)
        return cmpid, cmpname

    def create_coupling_inductors(self, compname, l1, l2, value=1, xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new coupling inductor.

        Parameters
        ----------
        compname : str
            Name of the coupling inductor.
        l1 : float, optional
            Value of inductor 1.
        l2 : float, optional
            Value of inductor 2.
        value : float, optional
            Value for the coupling inductor. The default is ``1``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.   

        Returns
        -------
        type
            ID of the coupling inductor.
        str
            Name of the coupling inductor.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Inductors", component_name="K_IND", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("Inductor1", l1)
        self.components[cmpid].set_property("Inductor2", l2)
        self.components[cmpid].set_property("CouplingFactor", value)
        return cmpid, cmpname

    @aedt_exception_handler
    def create_diode(self, compname=None,model_name="required", xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new diode.

        Parameters
        ----------
        compname : str
            Name of the diode. The default is ``None``.
        model_name : str, optional
            Name of the model. The default is ``"required"``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.
       
        Returns
        -------
        type
            ID of the diode.
        str
            Name of the diode.
        """
        cmpid, cmpname = self.create_component(compname,component_library="Diodes", component_name="DIODE_Level1", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)

        self.components[cmpid].set_property("MOD", model_name)
        return cmpid, cmpname

    @aedt_exception_handler
    def create_npn(self, compname=None, value=None, xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new transistor NPN.

        Parameters
        ----------
        compname : str
            Name of the transistor NPN. The default is ``None``.
        value : float, optional
            Value for the transistor NPN. The default is ``None``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.
        
        Returns
        -------
        type
            ID of the transistor NPN.
        str
            Name of the transistor NPN.
        """

        id, name = self.create_component(compname,component_library="BJTs", component_name="Level01_NPN", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)
        if value:
            self.components[id].set_property("MOD", value)
        return id, name

    @aedt_exception_handler
    def create_pnp(self, compname=None,value=50, xpos=0, ypos=0, angle=0, use_instance_id_netlist=False):
        """Create a new transistor PNP.

        Parameters
        ----------
        compname : str
            Name of the transistor PNP. The default is ``None``.
        value : float, optional
            Value for the transistor PPP. The default is ``None``.
        xpos : float, optional
            Position on the X axis. The default is ``0``.    
        ypos: float, optional
            Position on the Y axis. The default is ``0``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to use the instance ID in the net list. 
            The default is ``False``.
        
        Returns
        -------
        type
            ID of the transistor PNP.
        str
            Name of the transistor PNP.
        """
        id, name = self.create_component(compname, component_library="BJTs", component_name="Level01_PNP", xpos=xpos, ypos=ypos,
                                         angle=angle, use_instance_id_netlist=use_instance_id_netlist)
        if value:
            self.components[id].set_property( "MOD", value)

        return id, name

    @aedt_exception_handler
    def create_new_component_from_symbol(self,symbol_name, pin_lists, Refbase = "U", parameter_list=[], parameter_value=[]):
        """Create a new component from a symbol.

        Parameters
        ----------
        symbol_name : str
            Name of the symbol.
        pin_lists : list
            List of pin names.
        Refbase : str, optional
            Reference base. The default is ``"U"``.
        parameter_list : list
            List of parameters. The default is ``[]``.
        parameter_value : list
            List of parameter values. (The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        arg = ["NAME:" + symbol_name, "Info:=",
               ["Type:=", 0, "NumTerminals:=", len(pin_lists), "DataSource:=", "", "ModifiedOn:=", 1591858313, "Manufacturer:=", "",
                "Symbol:=", symbol_name, "ModelNames:=", "", "Footprint:=", "",
                "Description:=", "", "InfoTopic:=", "", "InfoHelpFile:=", "", "IconFile:=", "", "Library:=", "",
                "OriginalLocation:=", "Project", "IEEE:=", "", "Author:=", "", "OriginalAuthor:=", "",
                "CreationDate:=", 1591858278, "ExampleFile:=", "", "HiddenComponent:=", 0, "CircuitEnv:=", 0,
                "GroupID:=", 0], "CircuitEnv:=", 0, "Refbase:=", Refbase, "NumParts:=", 1, "ModSinceLib:=", True]

        for pin in pin_lists:
            arg.append("Terminal:=")
            arg.append([pin, pin, "A", False, 0, 1, "", "Electrical", "0"])
        arg.append("CompExtID:=")
        arg.append(1)
        arg2 = ["NAME:Parameters"]
        for el, val in zip(parameter_list, parameter_value):
            if type(val) is str:
                arg2.append("TextValueProp:=")
                arg2.append([el, "D", "", val])
            else:
                arg2.append("ValueProp:=")
                arg2.append([el, "D", "", val, False, ""])
        arg2.append("ButtonProp:=")
        arg2.append(["CosimDefinition", "D", "", "Edit", "Edit", 40501, "ButtonPropClientData:=", []])
        arg2.append("MenuProp:=")
        arg2.append(["CoSimulator", "D", "", "DefaultNetlist", 0])

        arg.append(arg2)
        spicesintax = Refbase + "@ID "
        id = 0
        while id < len(pin_lists):
            spicesintax += "%" + str(id) + " "
            id += 1
        for el, val in zip(parameter_list, parameter_value):
            if "MOD" in el:
                spicesintax += "@{} ".format(el)
            else:
                spicesintax += "{}=@{} ".format(el,val)

        arg3 = [
            "NAME:CosimDefinitions",
            [
                "NAME:CosimDefinition",
                "CosimulatorType:=", 4,
                "CosimDefName:=", "DefaultNetlist",
                "IsDefinition:=", True,
                "Connect:="	, True,
                "Data:="		, [					"Nexxim Circuit:="	, spicesintax],
                "GRef:="		, [					"Nexxim Circuit:="	, ""]
            ],
            "DefaultCosim:="	, "DefaultNetlist"
        ]
        arg.append(arg3)
        self.o_component_manager.Add(arg)
        return True

    @aedt_exception_handler
    def update_object_properties(self, o):
        """Update the object's properties.

        Parameters
        ----------
        o :
            Object to update.

        Returns
        -------
        type
            Object with properties.
        """
        name = o.composed_name
        proparray = self.oeditor.GetProperties("PassedParameterTab", name)
        for j in proparray:
            propval = retry_ntimes(10, self.oeditor.GetPropertyValue,"PassedParameterTab", name, j)
            o._add_property(j, propval)
        return o

    @aedt_exception_handler
    def get_comp_custom_settings(self, toolNum ,dc = 0, interp=0, extrap=1, conv=0, passivity=0, reciprocal="False", opt="", data_type=1):
        """

        Parameters
        ----------
        toolNum :
        
        dc :
            The default is ``0``.
        interp :
            The default is ``0``.
        extrap : 
            The default is ``1``.
        conv :
            The default is ``0``.
        passivity : optional
            The default is ``0``.
        reciprocal : bool, optional
            The default is ``False``.
        opt : str, optional
            The default is ``""``.
        data_type : optional
            Type of the data. The default is ``1``.  

        Returns
        -------
        type
            Custom settings for the resistor.
        """
        if toolNum == 1:
            custom = "NAME:DesignerCustomization"
        elif toolNum == 2:
            custom = "NAME:NexximCustomization"
        else:
            custom = "NAME:HSpiceCustomization"

        res = [custom,
               "DCOption:="		, dc,
               "InterpOption:="	, interp,
               "ExtrapOption:="	, extrap,
               "Convolution:="		, conv,
               "Passivity:="		, passivity,
               "Reciprocal:="		, reciprocal,
               "ModelOption:="		, opt,
               "DataType:="		, data_type]

        return res

    @aedt_exception_handler
    def add_subcircuit_hfss_link(self, compName, pin_names, source_project_path, source_project_name, source_design_name, solution_name = "Setup1 : Sweep"):
        """

        Parameters
        ----------
        compName : str
            Name of the subcircuit HFSS link.
        pin_names: list
            List of the pin names.
        source_project_path : str
            Path to the source project.      
        source_project_name: str
            Name  of the source project.
        source_design_name : str
            Name of the design.
        solution_name: str, optional
            Name of the solution and sweep. The 
            default is ``"Setup1 : Sweep"``.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        designer_customization = self.get_comp_custom_settings(1, 0, 0, 1, 0, 0, "False", "", 1)
        nexxim_customization = self.get_comp_custom_settings(2, 3, 1, 3, 0, 0, "False", "", 2)
        hspice_customization = self.get_comp_custom_settings(3, 1, 2, 3, 0, 0, "False", "", 3)

        compInfo = ["NAME:"+str(compName),
                    "Name:=", compName,
                    "ModTime:=", 1591855779,
                    "Library:=", "",
                    "LibLocation:=", "Project",
                    "ModelType:=", "hfss",
                    "Description:=", "",
                    "ImageFile:=", "",
                    "SymbolPinConfiguration:=", 0,
                    ["NAME:PortInfoBlk"],
                    ["NAME:PortOrderBlk"],
                    "DesignName:=", source_design_name,
                    "SolutionName:=", solution_name,
                    "NewToOldMap:=", [],
                    "OldToNewMap:=", [],
                    "PinNames:=", pin_names,
                    designer_customization,
                    nexxim_customization,
                    hspice_customization,
                    "NoiseModelOption:=", "External",
                    "WB_SystemID:=", "",
                    "IsWBModel:=", False,
                    "filename:=", source_project_path,
                    "numberofports:=", len(pin_names),
                    "Simulate:=", False,
                    "CloseProject:=", False,
                    "SaveProject:=", True,
                    "InterpY:=", True,
                    "InterpAlg:=", "auto",
                    "IgnoreDepVars:=", False,
                    "Renormalize:=", False,
                    "RenormImpedance:=", 50]

        self.o_model_manager.Add(compInfo)

        info = ["Type:=", 8, "NumTerminals:=", len(pin_names), "DataSource:=", "", "ModifiedOn:=", 1591855894, "Manufacturer:=", "",
                "Symbol:=", "", "ModelNames:=", "", "Footprint:=", "", "Description:=", "", "InfoTopic:=", "",
                "InfoHelpFile:=", "",	"IconFile:=", "hfss.bmp", "Library:=", "",	"OriginalLocation:=", "Project",
                "IEEE:=", "", "Author:=", "", "OriginalAuthor:=", "",	"CreationDate:=", 1591855894, "ExampleFile:=", "",
                "HiddenComponent:=", 0, "CircuitEnv:=", 0, "GroupID:=", 0]



        compInfo2 = ["NAME:"+str(compName),
                     "Info:=", info,
                     "CircuitEnv:=", 0,
                     "Refbase:=", "S",
                     "NumParts:=", 1,
                     "ModSinceLib:=", False]

        id = 0
        for pin in pin_names:
            compInfo2.append("Terminal:=")
            compInfo2.append([pin ,pin, "A", False, id, 1, "", "Electrical", "0"])
            id += 1

        compInfo2.append(["NAME:Properties","TextProp:=", ["Owner" ,"RD" ,"" ,"HFSS"]])
        compInfo2.append("CompExtID:=")
        compInfo2.append(5)
        compInfo2.append(["NAME:Parameters",
                          "TextProp:=", ["ModelName" ,"RD" ,"" ,"FieldSolver"],
                          "MenuProp:=", ["CoSimulator" ,"SD" ,"" ,"Default" ,0],
                          "ButtonProp:=", ["CosimDefinition" ,"SD" ,"" ,"Edit" ,"Edit" ,40501, "ButtonPropClientData:=", []]])
        compInfo2.append(["NAME:CosimDefinitions",
                          ["NAME:CosimDefinition",
                           "CosimulatorType:=", 103,
                           "CosimDefName:=", "Default",
                           "IsDefinition:=", True,
                           "Connect:=", True,
                           "ModelDefinitionName:=", compName,
                           "ShowRefPin2:=", 2,
                           "LenPropName:=", ""],
                          "DefaultCosim:=", "Default"])

        self.o_component_manager.Add(compInfo2)
        self._parent.odesign.AddCompInstance(compName)
        self.refresh_all_ids()
        for el in self.components:
            item = compName
            item2 = self.components[el].composed_name
            if compName in self.components[el].composed_name:
                return el, self.components[el].composed_name
        return False

    @aedt_exception_handler
    def set_sim_option_on_hfss_subcircuit(self, component):
        """

        Parameters
        ----------
        component : str
            Address of the component instance. For example, ``"Inst@Galileo_cutout3;87;1"``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        complist = component.split(";")
        complist2 = complist[0].split("@")
        arg = ["NAME:AllTabs"]
        arg1 = ["NAME:Model"]
        arg2 = ["NAME:PropServers", "Component@"+str(complist2[1])]
        arg3 = ["NAME:ChangedProps",["NAME:Simulation option", "Value:=", "Simulate missing solutions"]]

        arg1.append(arg2)
        arg1.append(arg3)
        arg.append(arg1)

        self._parent._oproject.ChangeProperty(arg)

        argo = ["NAME:AllTabs", ["NAME:Component", ["NAME:PropServers", str(component)]]]

        self.oeditor.ChangeProperty(argo)

        pass

    @aedt_exception_handler
    def set_sim_solution_on_hfss_subcircuit(self, component, solution_name="Setup1 : Sweep"):
        """

        Parameters
        ----------
        component : str
            Address of the component instance. For example, ``"Inst@Galileo_cutout3;87;1"``.
        solution_name: str, optional
            Name of the solution and sweep. The
            default is ``"Setup1 : Sweep"``.

        Returns
        -------
        bool
        """
        complist = component.split(";")
        complist2 = complist[0].split("@")
        arg = ["NAME:AllTabs"]
        arg1 = ["NAME:Model"]
        arg2 = ["NAME:PropServers", "Component@" + str(complist2[1])]
        arg3 = ["NAME:ChangedProps", ["NAME:Solution", "Value:=", solution_name]]

        arg1.append(arg2)
        arg1.append(arg3)
        arg.append(arg1)

        self._parent._oproject.ChangeProperty(arg)

        argo = ["NAME:AllTabs", ["NAME:Component", ["NAME:PropServers", str(component)]]]

        self.oeditor.ChangeProperty(argo)

        return True

    @aedt_exception_handler
    def refresh_dynamic_link(self, component_name):
        """

        Parameters
        ----------
        component_name : str
            Name of the dynamic link.
            

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        self.o_component_manager.UpdateDynamicLink(component_name)

    @aedt_exception_handler
    def push_excitations(self, instance_name, thevenin_calculation=False, setup_name="LinearFrequency"):
        """

        Parameters
        ----------
        instance_name : str
            Name of the instance.
        thevenin_calculation : bool, optional
            Whether to perform the Thevenin equivalent calculation. The default is ``False``.
        setup_name : str
            Name of the Solution Setup to push

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        arg = ["NAME:options",
               "CalcThevenin:=", thevenin_calculation,
               "Sol:=", setup_name]

        self.oeditor.PushExcitations(instance_name, arg)
        pass

    @aedt_exception_handler
    def assign_sin_excitation2ports(self, ports, settings):
        """

        Parameters
        ----------
        ports : list
            List of circuit ports to assign to the excitation.
        settings : list
            List of parameter values to use in voltage sinusoidal excitation creation.
            
            Valures are given in this order:
        
            0: AC magnitude for small-signal analysis
            
            1: AC phase for small-signal analysis
            
            2: DC voltage
            
            3: Voltage offset from zero
            
            4: Voltage amplitude
            
            5: Frequency
            
            6: Delay to start of sine wave
            
            7: Damping factor
            
            8: Phase delay
            
            9: Frequency to use for harmonic balance analysis

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        id = self.create_unique_id()

        arg1 = ["NAME:NexximSources",
                ["NAME:NexximSources",
                 ["NAME:Data",
                  ["NAME:VoltageSinusoidal"+str(id),
                   "DataId:=", "Source"+str(id),
                   "Type:=", 1,
                   "Output:=", 0,
                   "NumPins:=", 2,
                   "Netlist:=", "V@ID %0 %1 *DC(DC=@DC) SIN(?VO(@VO) ?VA(@VA) ?FREQ(@FREQ) ?TD(@TD) ?ALPHA(@ALPHA) ?THETA(@THETA)) *TONE(TONE=@TONE) *ACMAG(AC @ACMAG @ACPHASE)",
                   "CompName:=", "Nexxim Circuit Elements\\Independent Sources:V_SIN",
                   "FDSFileName:=", "",
                   ["NAME:Properties",
                    "TextProp:=", ["LabelID","HD","Property string for netlist ID","V@ID"],
                    "ValueProp:=", ["ACMAG","OD","AC magnitude for small-signal analysis (Volts)",settings[0],0],
                    "ValuePropNU:=", ["ACPHASE","OD","AC phase for small-signal analysis",settings[1],0,"deg"],
                    "ValueProp:=", ["DC","OD","DC voltage (Volts)",settings[2],0],
                    "ValueProp:=", ["VO","OD","Voltage offset from zero (Volts)",settings[3],0],
                    "ValueProp:=", ["VA","OD","Voltage amplitude (Volts)",settings[4],0],
                    "ValueProp:=", ["FREQ","OD","Frequency (Hz)",settings[5],0],
                    "ValueProp:=", ["TD","OD","Delay to start of sine wave (seconds)",settings[6],0],
                    "ValueProp:=", ["ALPHA","OD","Damping factor (1/seconds)",settings[7],0],
                    "ValuePropNU:=", ["THETA","OD","Phase delay",settings[8],0,"deg"],
                    "ValueProp:=", ["TONE","OD","Frequency (Hz) to use for harmonic balance analysis, should be a submultiple of (or equal to) the driving frequency and should also be included in the HB analysis setup",settings[9],0],
                    "TextProp:=", ["ModelName","SHD","","V_SIN"],
                    "MenuProp:=", ["CoSimulator","D","","DefaultNetlist",0],
                    "ButtonProp:=", ["CosimDefinition","D","","","Edit",40501, "ButtonPropClientData:=", [] ] ] ] ] ] ]

        arg2 = ["NAME:ComponentConfigurationData"]

        arg3 = ["NAME:ComponentConfigurationData", ["NAME:EnabledPorts", "VoltageSinusoidal"+str(id)+":=", ports],
                ["NAME:EnabledMultipleComponents", "VoltageSinusoidal"+str(id)+":=", [] ] ]

        for prt in ports:
            arg_temp = []
            arg_temp = ["NAME:EnabledAnalyses", ["NAME:"+str(prt), str(prt)+":=", [] ], ["NAME:VoltageSinusoidal"+str(id), str(prt)+":=", [] ] ]
            arg3.append(arg_temp)

        arg2.append(arg3)

        self._parent.odesign.UpdateSources(arg1,arg2)

        pass



"""
This module contains these classes: ''Layer'' and ''Layers''.  

This module provides all layer stackup functionalities for the Circuit and HFSS 3D Layout tools.
"""
from __future__ import absolute_import
import random
import string
import sys
from collections import defaultdict
from ..generic.general_methods import aedt_exception_handler

@aedt_exception_handler
def str2bool(str0):
    """Convert a string to a Boolean value.

    Parameters
    ----------
    str0 : str
       String to convert.   

    Returns
    -------
    bool
        ``True`` when successful, ``False`` when failed.
    """
    if str0.lower() == "false":
        return False
    elif str0 == "true":
        return True
    else:
        return ""

def conv_number(number, typen=float):
    """Convert a number.

    Parameters
    ----------
    float
       Number represented a float.
        
    typen : 
         The default is ``float``.

    Returns
    -------

    """
    if typen is float:
        try:
            return float(number)
        except:
            return number
    elif typen is int:
        try:
            return int(number)
        except:
            return number


@aedt_exception_handler
def getIfromRGB(rgb):
    """

    Parameters
    ----------
    rgb :
        

    Returns
    -------

    """
    red = rgb[2]
    green = rgb[1]
    blue = rgb[0]
    RGBint = (red << 16) + (green << 8) + blue
    return RGBint


class Layer(object):
    """Layer class.
    
    Parameters
    ----------
    editor :
    
    type : string, optional
        The default is ``"signal"``.
    layerunits : str, optional
        The default is ``"mm"``.
    roughunit: str, optional
        The default is ``"um"``.
    negative: bool, optional
        Whether the geometry on the layer is cut away 
        from the layer. The default is ``False``.
    """

    @property
    def visflag(self):
        """Visibility flag for objects on the layer. """
        visflag = 0
        if not self.IsVisible:
            visflag = 0
        else:
            if self.IsMeshBackgroundMaterial:
                visflag += 64
            if self.IsMeshOverlay:
                visflag += 32
            if self.IsVisibleShape:
                visflag += 1
            if self.IsVisiblePath:
                visflag += 2
            if self.IsVisiblePad:
                visflag += 4
            if self.IsVisibleHole:
                visflag += 8
            if self.IsVisibleComponent:
                visflag += 16
        return visflag

    def __init__(self, editor, type="signal", layerunits="mm", roughunit="um", negative=False):
        self.LengthUnit = layerunits
        self.LengthUnitRough = roughunit
        self.oeditor = editor
        self.name = None
        self.type = type
        self.id = 0
        self.color = 8026109
        self.transparency = 0.3
        self.IsVisible = True
        self.IsVisibleShape = True
        self.IsVisiblePath = True
        self.IsVisiblePad = True
        self.IsVisibleHole = True
        self.IsVisibleComponent = True
        self.IsMeshBackgroundMaterial = True
        self.IsMeshOverlay = True
        self.locked = False
        self.topbottom = "neither"
        self.pattern = 1
        self.drawoverride = 0
        self.regions = []
        self.thickness = 0
        self.lowerelevation = 0
        self.roughness = 0
        self.botroughness = 0
        self.toprounghenss = 0
        self.sideroughness = 0
        self.material = "copper"
        self.fillmaterial = "FR4_epoxy"
        self.etch = 0
        self.usp = True
        self.useetch = True
        self.user = True
        self.RMdl = "Huray"
        self.NR = 0.5
        self.HRatio = 2.9
        self.BRMdl = "Huray"
        self.BNR = 0.5
        self.BHRatio = 2.9
        self.SRMdl = "Huray"
        self.SNR = 0.5
        self.SHRatio = 2.9
        self.hfssSp = {"Sn": "HFSS", "Sv": None}
        self.planaremSp = {"Sn": "PlanarEM", "Sv": None}
        self.index = 1
        self.IsNegative = negative

    @aedt_exception_handler
    def set_layer_color(self, r, g, b):
        """Update the color of the layer.

        Parameters
        ----------
        r : int
            Red color value.
        g : int
            Green color value.
        b :  int
            Blue color value.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        rgb = [r, g, b]
        self.color = getIfromRGB(rgb)
        self.update_stackup_layer()
        return True

    @aedt_exception_handler
    def create_stackup_layer(self):
        """Create a stackup layer.
        
        Parameters
        ----------

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        self.remove_stackup_layer()
        if self.type == "signal":
            self.oeditor.AddStackupLayer([
                "NAME:stackup layer",
                "Name:=", self.name,
                "Type:=", self.type,
                "Top Bottom:=", self.topbottom,
                "Color:=", self.color,
                "Transparency:=", self.transparency,
                "Pattern:=", self.pattern,
                "VisFlag:=", self.visflag,
                "Locked:=", self.locked,
                "DrawOverride:=", self.drawoverride,
                "Regions:=", self.regions,
                [
                    "NAME:Sublayer",
                    "Thickness:=", self.thickness,
                    "LowerElevation:=", self.lowerelevation,
                    "Roughness:=", self.arg_with_dim(self.roughness, self.LengthUnitRough),
                    "BotRoughness:=", self.arg_with_dim(self.botroughness, self.LengthUnitRough),
                    "SideRoughness:=", self.arg_with_dim(self.toprounghenss, self.LengthUnitRough),
                    "Material:=", self.arg_with_dim(self.material),
                    "FillMaterial:=", self.arg_with_dim(self.fillmaterial)
                ],
                "Neg:="		, self.IsNegative,
                "Usp:=", self.usp,
                "Etch:=", str(self.etch),
                "UseEtch:=", self.useetch,
                "UseR:=", self.user,
                "RMdl:=", self.RMdl,
                "NR:=", self.arg_with_dim(self.NR, self.LengthUnitRough),
                "HRatio:=", str(self.HRatio),
                "BRMdl:=", self.BRMdl,
                "BNR:=", self.arg_with_dim(self.BNR, self.LengthUnitRough),
                "BHRatio:=", str(self.BHRatio),
                "SRMdl:=", self.SRMdl,
                "SNR:=", self.arg_with_dim(self.SNR, self.LengthUnitRough),
                "SHRatio:=", str(self.SHRatio)
            ])
        else:
            self.oeditor.AddStackupLayer([
                "NAME:stackup layer",
                "Name:=", self.name,
                "Type:=", self.type,
                "Top Bottom:=", self.topbottom,
                "Color:=", self.color,
                "Transparency:=", self.transparency,
                "Pattern:=", self.pattern,
                "VisFlag:=", self.visflag,
                "Locked:=", self.locked,
                "DrawOverride:=", self.drawoverride,
                "Regions:=", self.regions,
                [
                    "NAME:Sublayer",
                    "Thickness:=", self.thickness,
                    "LowerElevation:=", self.lowerelevation,
                    "Roughness:=", 0,
                    "BotRoughness:=", 0,
                    "SideRoughness:=", 0,
                    "Material:=", self.arg_with_dim(self.material),
                    "FillMaterial:=", ""
                ]
            ])
        infos = self.oeditor.GetLayerInfo(self.name)
        infos = [i.split(": ") for i in infos]
        infosdict = {i[0]: i[1] for i in infos}
        self.id = int(infosdict["LayerId"])
        return True

    @aedt_exception_handler
    def arg_with_dim(self, Value, sUnits=None):
        """

        Parameters
        ----------
        Value :
            
        sUnits :
             The default is ``None``.

        Returns
        -------

        """
        if type(Value) is str:
            val = Value
        else:
            if sUnits is None:
                sUnits = self.LengthUnit
            val = "{0}{1}".format(Value, sUnits)

        return val

    @aedt_exception_handler
    def update_stackup_layer(self):
        """Update the stackup layer.
        
        .. note::
           This method is valid for release 2021 R1 and later.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        if self.type == "signal":
            self.oeditor.ChangeLayer([
                "NAME:SLayer",
                "Name:=", self.name,
                "ID:=", self.id,
                "Type:=", self.type,
                "Top Bottom:=", self.topbottom,
                "Color:=", self.color,
                "Transparency:=", self.transparency,
                "Pattern:=", self.pattern,
                "VisFlag:=", self.visflag,
                "Locked:=", self.locked,
                "DrawOverride:=", self.drawoverride,
                "Zones:="	, [0],
                [
                    "NAME:Sublayer",
                    "Thickness:=", self.thickness,
                    "LowerElevation:=", self.lowerelevation,
                    "Roughness:=", self.arg_with_dim(self.roughness, self.LengthUnitRough),
                    "BotRoughness:=", self.arg_with_dim(self.botroughness, self.LengthUnitRough),
                    "SideRoughness:=", self.arg_with_dim(self.toprounghenss, self.LengthUnitRough),
                    "Material:=", self.material,
                    "FillMaterial:=", self.fillmaterial
                ],
                "Usp:=", self.usp,
                "Etch:=", str(self.etch),
                "UseEtch:=", self.useetch,
                "UseR:=", self.user,
                "RMdl:=", self.RMdl,
                "NR:=", self.arg_with_dim(self.NR, self.LengthUnitRough),
                "HRatio:=", str(self.HRatio),
                "BRMdl:=", self.BRMdl,
                "BNR:=", self.arg_with_dim(self.BNR, self.LengthUnitRough),
                "BHRatio:=", str(self.BHRatio),
                "SRMdl:=", self.SRMdl,
                "SNR:=", self.arg_with_dim(self.SNR, self.LengthUnitRough),
                "SHRatio:=", str(self.SHRatio)
            ])
        else:
            self.oeditor.ChangeLayer([
                "NAME:stackup layer",
                "Name:=", self.name,
                "ID:=", self.id,
                "Type:=", self.type,
                "Top Bottom:=", self.topbottom,
                "Color:=", self.color,
                "Transparency:=", self.transparency,
                "Pattern:=", self.pattern,
                "VisFlag:=", self.visflag,
                "Locked:=", self.locked,
                "DrawOverride:=", self.drawoverride,
                "Regions:=", self.regions,
                [
                    "NAME:Sublayer",
                    "Thickness:=", self.thickness,
                    "LowerElevation:=", self.lowerelevation,
                    "Roughness:=", 0,
                    "BotRoughness:=", 0,
                    "SideRoughness:=", 0,
                    "Material:=", self.material,
                    "FillMaterial:=", ""
                ]
            ])
        return True


    @aedt_exception_handler
    def remove_stackup_layer(self):
        """Remove the stackup layer.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        if self.name in self.oeditor.GetStackupLayerNames():
            self.oeditor.RemoveLayer(self.name)
            return True
        return False


class Layers(object):
    """Layers class.
    
    Parameters
    ----------
    parent :
    
    modeler :
    
    roughnessunits : str, optional
       Units for the roughness of layers. The default is ``"um"``.
    
    """

    @property
    def _messenger(self):
        """_messenger."""
        return self._parent._messenger

    @property
    def modeler(self):
        """Modeler."""
        return self._modeler

    @property
    def oeditor(self):
        """Editor object."""
        return self.modeler.oeditor

    @property
    def LengthUnit(self):
        """Length units."""
        return self.modeler.model_units

    @property
    def all_layers(self):
        """Stackup layers.
        
        Returns
        -------
        list
           List of stackup layers.
        """
        return self.oeditor.GetStackupLayerNames()

    @property
    def all_signal_layers(self):
        """Signal layers.
        
        Returns
        -------
        list
            List of signal layers.
        """
        a = self.all_layers
        sig = []
        for lay in a:
            layid = self.layer_id(lay)
            if layid not in self.layers:
                self.refresh_all_layers()
            if self.layers[layid].type == "signal":
                sig.append(lay)
        return sig

    @property
    def all_diel_layers(self):
        """Dielectric layers.
        
        Returns
        -------
        list
            List of dielectric layers.
        """
        a = self.all_layers
        die = []
        for lay in a:
            layid = self.layer_id(lay)

            if layid not in self.layers:
                self.refresh_all_layers()
            if self.layers[layid].type == "dielectric":
                die.append(layid)
        return die

    def __init__(self, parent, modeler, roughnessunits="um"):
        self._modeler = modeler
        self._parent = parent
        self._currentId = 0
        self.layers = defaultdict(Layer)
        self.lengthUnitRough = roughnessunits

    @aedt_exception_handler
    def layer_id(self, name):
        """Retrieve a layer ID.

        Parameters
        ----------
        name :  str
            Name of the layer. 

        Returns
        -------
        type
            Layer ID if the layer name exists.
        """
        for el in self.layers:
            if self.layers[el].name == name:
                return el
        return None

    @aedt_exception_handler
    def refresh_all_layers(self):
        """Refresh all layers in the current stackup.
       
        Returns
        -------
        int
            Number of layers in the current stackup.
        """
        layernames = self.oeditor.GetStackupLayerNames()
        for el in layernames:
            o = Layer(self.oeditor, "signal", self.LengthUnit, self.lengthUnitRough)
            o.name = el
            infos = self.oeditor.GetLayerInfo(el)
            infos = [i.split(": ") for i in infos]
            infosdict = {i[0]: i[1] for i in infos}
            o.type = infosdict["Type"]
            o.locked = str2bool(infosdict["IsLocked"])
            o.id = int(infosdict["LayerId"])
            o.topbottom = infosdict["TopBottomAssociation"]
            o.IsVisible = infosdict["IsVisible"]
            if "IsVisiblePath" in infosdict:
                o.IsVisiblePath = infosdict["IsVisiblePath"]
                o.IsVisiblePad = infosdict["IsVisiblePad"]
                o.IsVisibleComponent = infosdict["IsVisibleComponent"]
                o.IsVisibleShape = infosdict["IsVisibleShape"]
                o.IsVisibleHole = infosdict["IsVisibleHole"]
            o.color = int(infosdict["Color"][:-1])
            o.index = int(infosdict["Index"])
            o.thickness = conv_number(infosdict["LayerThickness"])
            o.lowerelevation = conv_number(infosdict["LowerElevation0"])
            o.fillmaterial = infosdict['FillMaterial0']
            o.material = infosdict["Material0"]
            if "Roughness0 Type" in infosdict:
                o.RMdl = infosdict["Roughness0 Type"]
                o.NR = infosdict["Roughness0"].split(", ")[0]
                o.HRatio = conv_number(infosdict["Roughness0"].split(", ")[1])
            if "BottomRoughness0 Type" in infosdict:
                o.BRMdl = infosdict["BottomRoughness0 Type"]
                o.BNR = infosdict["BottomRoughness0"].split(", ")[0]
                o.BHRatio = conv_number(infosdict["BottomRoughness0"].split(", ")[1])
            if "SideRoughness0 Type" in infosdict:
                o.SRMdl = infosdict["SideRoughness0 Type"]
                o.SNR = infosdict["SideRoughness0"].split(", ")[0]
                o.SHRatio = conv_number(infosdict["SideRoughness0"].split(", ")[1])
            self.layers[o.id] = o
        return len(self.layers)

    @aedt_exception_handler
    def add_layer(self, layername, layertype="signal", thickness="0mm", elevation="0mm", material="copper", isnegative=False):
        """Add a layer.

        Parameters
        ----------
        layername : str
            Name of the layer.
        layertype : str, optional
            Type of the layer. The default is ``"signal"``.
        thickness : str, optional
            Thickness with units. The default is ``"0mm"``.
        elevation : str, optional
            Elevation with units. The default is ``"0mm"``.
        material : str, optional
            Name of the material. The default is ``"copper"``.

        Returns
        -------
        type
            Layer object.
        """
        newlayer = Layer(self.oeditor, layertype, self.LengthUnit, self.lengthUnitRough, isnegative)
        newlayer.name = layername
        newlayer.thickness = thickness
        newlayer.lowerelevation = elevation
        newlayer.material = material
        newlayer.create_stackup_layer()
        self.layers[newlayer.id] = newlayer
        return self.layers[newlayer.id]



"""
This module contains the ``Edb3DLayout`` class.
"""

import warnings
import os
from .general import *
from ..generic.general_methods import get_filename_without_extension, generate_unique_name

try:
    import clr
    from System import Convert, String
    from System import Double, Array
    from System.Collections.Generic import List
except ImportError:
    warnings.warn('This module requires pythonnet.')


class Edb3DLayout(object):
    """Edb3DLayout class."""

    def __init__(self, parent):
        self.parent = parent

    @property
    def _hfss_terminals(self):
        return self.parent.edblib.HFSS3DLayout.HFSSTerminalMethods

    @property
    def _hfss_ic_methods(self):
        return self.parent.edblib.HFSS3DLayout.ICMethods

    @property
    def _hfss_setup(self):
        return self.parent.edblib.HFSS3DLayout.HFSSSetup

    @property
    def _hfss_mesh_setup(self):
        return self.parent.edblib.HFSS3DLayout.Meshing

    @property
    def _sweep_methods(self):
        return self.parent.edblib.SimulationSetup.SweepMethods

    @property
    def _edb(self):
        return self.parent.edb

    @property
    def _active_layout(self):
        return self.parent.active_layout

    @property
    def _cell(self):
        return self.parent.cell

    @property
    def _db(self):
        return self.parent.db

    @property
    def _builder(self):
        return self.parent.builder

    @aedt_exception_handler
    def get_trace_width_for_traces_with_ports(self):
        """Retrieve the trace width for traces with ports.

        Returns
        -------
        dict
            Dictionary of trace width data.
        """
        mesh =  self._hfss_mesh_setup.GetMeshOperation(self._builder)
        if mesh.Item1:
            return convert_netdict_to_pydict(mesh.Item2)
        else:
            return {}

    @aedt_exception_handler
    def create_circuit_port(self, pin_list):
        """Create a circuit port.

        Parameters
        ----------
        pin_list : list
            Pin object or a pin list object.          

        Returns
        -------
        list
           List of names of the created ports.
        
        """
        pinList = convert_py_list_to_net_list(pin_list)
        if self._hfss_terminals.CreateHfssPorts(self._builder, pinList):
            return True

    @aedt_exception_handler
    def create_coax_port_on_component(self, ref_des_list, net_list):
        """Create a coaxial port on a component or component list on a net or net list.
        
        .. note::
           The name of the new coaxial port is automatically assigned.

        Parameters
        ----------
        ref_des_list : list, str
            List of one or more reference designators.
            
        net_list : list, str
            List of one or more nets.  

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        coax = []
        if not isinstance(ref_des_list, list):
            ref_des_list = [ref_des_list]
        if not isinstance(net_list, list):
            ref_des_list = [net_list]
        for ref in ref_des_list:
            for pinname, pin in self.parent.core_components.components[ref].pins.items():
                if pin.net in net_list and pin.pin.IsLayoutPin():
                   port_name = "{}_{}_{}".format(ref,pin.net,pin.pin.GetName())
                   if "IronPython" in sys.version or ".NETFramework" in sys.version:
                       res, fromLayer_pos, toLayer_pos = pin.pin.GetLayerRange()
                   else:
                       res, fromLayer_pos, toLayer_pos = pin.pin.GetLayerRange(None, None)
                   if self._edb.Cell.Terminal.PadstackInstanceTerminal.Create(self._active_layout, pin.pin.GetNet(), port_name, pin.pin, toLayer_pos):
                       coax.append(port_name)
        return coax

    @aedt_exception_handler
    def create_hfss_ports_on_padstack(self, pinpos, portname=None):
        """Create a HFSS port on a padstack.

        Parameters
        ----------
        pinpos : 
            Position of the pin.
            
        portname : str, optional
             Name of the port. The default is ``None``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        if "IronPython" in sys.version or ".NETFramework" in sys.version:
            res, fromLayer_pos, toLayer_pos = pinpos.GetLayerRange()
        else:
            res, fromLayer_pos, toLayer_pos = pinpos.GetLayerRange(None, None)
        if not portname:
            portname = generate_unique_name("Port_" + pinpos.GetNet().GetName())
        edbpointTerm_pos = self._edb.Cell.Terminal.PadstackInstanceTerminal.Create(self._active_layout, pinpos.GetNet(),
                                                                          portname, pinpos,
                                                                          toLayer_pos)
        if edbpointTerm_pos:
            return True
        else:
            return False





# -*- coding: utf-8 -*-
"""This module contains the `Circuit` class."""

from __future__ import absolute_import
import math
from .application.AnalysisNexxim import FieldAnalysisCircuit
from .desktop import exception_to_desktop
from .generic.general_methods import generate_unique_name, aedt_exception_handler
import re
import os

RKM_MAPS = {
    # Resistors
    'L': 'm',
    'R': '',
    'E': '',
    'k': 'k',
    'K': 'k',
    'M': 'M',
    'G': 'G',
    'T': 'T',
    # Capacitors/Inductors
    'F': '',
    'H': '',
    'h': '',
    'm': 'm',
    'u': 'μ',
    'μ': 'μ',
    'U': 'μ',
    'n': 'n',
    'N': 'n',
    'p': 'p',
    'P': 'p',
    'mF': 'm',
    'uF': 'μ',
    'μF': 'μ',
    'UF': 'μ',
    'nF': 'n',
    'NF': 'n',
    'pF': 'p',
    'PF': 'p',
    'mH': 'm',
    'uH': 'μ',
    'μH': 'μ',
    'UH': 'μ',
    'nH': 'n',
    'NH': 'n',
    'pH': 'p',
    'PH': 'p',
}

AEDT_MAPS = {
    'μ': 'u'
}


def from_rkm(code):
    """Convert an RKM code string to a string with a decimal point.

    Parameters
    ----------
    code : str
        RKM code string.

    Returns
    -------
    str
        String with a decimal point and an R value.

    Examples
    --------
    >>> from pyaedt.circuit import from_rkm
    >>> from_rkm('R47')
    '0.47'

    >>> from_rkm('4R7')
    '4.7'

    >>> from_rkm('470R')
    '470'

    >>> from_rkm('4K7')
    '4.7k'

    >>> from_rkm('47K')
    '47k'

    >>> from_rkm('47K3')
    '47.3k'

    >>> from_rkm('470K')
    '470k'

    >>> from_rkm('4M7')
    '4.7M'

    """

    # Matches RKM codes that start with a digit.
    # fd_pattern = r'([0-9]+)([LREkKMGTFmuµUnNpP]+)([0-9]*)'
    fd_pattern = r'([0-9]+)([{}]+)([0-9]*)'.format(''.join(RKM_MAPS.keys()), )
    # matches rkm codes that end with a digit
    # ld_pattern = r'([0-9]*)([LREkKMGTFmuµUnNpP]+)([0-9]+)'
    ld_pattern = r'([0-9]*)([{}]+)([0-9]+)'.format(''.join(RKM_MAPS.keys()))

    fd_regex = re.compile(fd_pattern, re.I)
    ld_regex = re.compile(ld_pattern, re.I)

    for regex in [fd_regex, ld_regex]:
        m = regex.match(code)
        if m:
            fd, base, ld = m.groups()
            ps = RKM_MAPS[base]

            if ld:
                return_str = ''.join([fd, '.', ld, ps])
            else:
                return_str = ''.join([fd, ps])
            return return_str
    return code


def to_aedt(code):
    """

    Parameters
    ----------
    code : str
    
    Returns
    -------
    str
    
    """
    pattern = r'([{}]{})'.format(''.join(AEDT_MAPS.keys()), '{1}')
    regex = re.compile(pattern, re.I)
    return_code = regex.sub(lambda m: AEDT_MAPS.get(m.group(), m.group()), code)
    return return_code


def from_rkm_to_aedt(code):
    """

    Parameters
    ----------
    code : str
        

    Returns
    -------
    str
    
    """
    return to_aedt(from_rkm(code))


class Circuit(FieldAnalysisCircuit, object):
    """Circuit application interface.

    Parameters
    ----------
    projectname : str, optional
        Name of the project to select or the full path to the project
        or AEDTZ archive to open.  The default is ``None``, in which
        case an attempt is made to get an active project. If no
        projects are present, an empty project is created.
    designname : str, optional
        Name of the design to select. The default is ``None``, in
        which case an attempt is made to get an active design. If no
        designs are present, an empty design is created.
    solution_type : str, optional
        Solution type to apply to the design. The default is
        ``None``, which applies the default type.
    setup_name : str, optional
        Name of the setup to use as the nominal. The default is
        ``None``, in which case the active setup is used or
        nothing is used.
    specified_version: str, optional
        Version of AEDT to use. The default is ``None``, in which case
        the active version or latest installed version is  used.
    NG : bool, optional
        Whether to run AEDT in the non-graphical mode. The default
        is``False``, which launches AEDT in the graphical mode.
    AlwaysNew : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the
        machine.  The default is ``True``.
    release_on_exit : bool, optional
        Whether to release AEDT on exit.
    student_version : bool, optional
        Whether to open the AEDT student version. The default is ``False``.

    Examples
    --------
    Create an instance of Circuit and connect to an existing HFSS
    design or create a new HFSS design if one does not exist.

    >>> from pyaedt import Circuit
    >>> aedtapp = Circuit()

    Create an instance of Circuit and link to a project named
    ``projectname``. If this project does not exist, create one with
    this name.

    >>> aedtapp = Circuit(projectname)

    Create an instance of Circuit and link to a design named
    ``designname`` in a project named ``projectname``.

    >>> aedtapp = Circuit(projectname,designame)

    Create an instance of Circuit and open the specified project,
    which is ``myfie.aedt``.

    >>> aedtapp = Circuit("myfile.aedt")

    Create an instance of Circuit using the 2021 R1 version and
    open the specified project, which is ``myfie.aedt``.

    >>> aedtapp = Circuit(specified_version="2021.1", projectname="myfile.aedt")

    Create an instance of Circuit using the 2021 R2 student version and open
    the specified project, which is named ``"myfile.aedt"``.

    >>> hfss = Circuit(specified_version="2021.2", projectname="myfile.aedt", student_version=True)

    Examples
    --------
    Create an instance of Circuit and connect to an existing HFSS
    design or create a new HFSS design if one does not exist.

    >>> from pyaedt import Circuit
    >>> aedtapp = Circuit()

    Create an instance of Circuit and link to a project named
    ``projectname``. If this project does not exist, create one with
    this name.

    >>> aedtapp = Circuit(projectname)

    Create an instance of Circuit and link to a design named
    ``designname`` in a project named ``projectname``.

    >>> aedtapp = Circuit(projectname,designame)

    Create an instance of Circuit and open the specified project,
    which is ``myfie.aedt``.

    >>> aedtapp = Circuit("myfile.aedt")

    Create an instance of Circuit using the 2021 R1 version and
    open the specified project, which is ``myfie.aedt``.

    >>> aedtapp = Circuit(specified_version="2021.1", projectname="myfile.aedt")

    """
    def __init__(self, projectname=None, designname=None, solution_type=None, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False, student_version=False):
        FieldAnalysisCircuit.__init__(self, "Circuit Design", projectname, designname, solution_type, setup_name,
                                      specified_version, NG, AlwaysNew, release_on_exit, student_version)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        if ex_type:
            exception_to_desktop(self, ex_value, ex_traceback)

    @property
    def onetwork_data_explorer(self):
        return self._desktop.GetTool("NdExplorer")

    @aedt_exception_handler
    def _get_number_from_string(self, stringval):
        value = stringval[stringval.find("=") + 1:].strip().replace("{", "").replace("}", "").replace(",", ".")
        try:
            float(value)
            return value
        except:
            return from_rkm_to_aedt(value)

    @aedt_exception_handler
    def create_schematic_from_netlist(self, file_to_import):
        """Create a circuit schematic from an HSpice netlist.

        Supported currently:
        
        * R
        * L
        * C
        * Diodes
        * Bjts
        * Discrete components with syntax ``Uxxx net1 net2 ... netn modname``

        Parameters
        ----------
        file_to_import : str
            Full path to the HSpice file to import.

        Returns
        -------
        bool
             ``True`` when successful, ``False`` when failed.
        
        """
        xpos = 0
        ypos = 0
        delta = 0.0508
        use_instance = True
        model = []
        self._desktop.CloseAllWindows()
        autosave = False
        if self._desktop.GetAutoSaveEnabled() == 1:
            self._desktop.EnableAutoSave(False)
            autosave = True
        with open(file_to_import, 'rb') as f:
            for line in f:
                line = line.decode('utf-8')
                if ".param" in line[:7].lower():
                    try:
                        ppar = line[7:].split("=")[0]
                        pval = line[7:].split("=")[1]
                        self[ppar] = pval
                        xpos = 0.0254
                    except:
                        pass
                elif ".model" in line[:7].lower() or ".lib" in line[:4].lower():
                    model.append(line)
        if model:
            self.modeler.components.create_symbol("Models_Netlist", [])
            self.modeler.components.create_new_component_from_symbol("Models_Netlist", [], "")
            self.modeler.components.create_component(None, component_library=None, component_name="Models_Netlist",
                                                     xpos=xpos, ypos=0, global_netlist_list=model)
            self.modeler.components.disable_data_netlist(component_name="Models_Netlist")
            xpos += 0.0254
        counter = 0
        with open(file_to_import, 'rb') as f:
            for line in f:
                line = line.decode('utf-8')
                mycomp = None
                fields = line.split(" ")
                name = fields[0].replace(".", "")

                if fields[0][0] == "R":
                    if "{" in fields[3][0]:
                        value = fields[3].strip()[1:-1]
                    elif '/' in fields[3] and '"' not in fields[3][0] and "'" not in fields[3][0] and "{" not in \
                            fields[3][0]:
                        value = self._get_number_from_string(fields[3].split('/')[0])
                    else:
                        value = self._get_number_from_string(fields[3])
                    mycomp, mycompname = self.modeler.components.create_resistor(name, value, xpos, ypos,
                                                                                 use_instance_id_netlist=use_instance)
                elif fields[0][0] == "L":
                    if len(fields) > 4 and "=" not in fields[4]:
                        try:
                            float(fields[4])
                        except:
                            self._messenger.add_warning_message(
                                "Component {} Not Imported. Check it and manually import".format(name))
                            continue
                    if "{" in fields[3][0]:
                        value = fields[3].strip()[1:-1]
                    elif '/' in fields[3] and '"' not in fields[3][0] and "'" not in fields[3][0]:
                        value = self._get_number_from_string(fields[3].split('/')[0])
                    else:
                        value = self._get_number_from_string(fields[3])
                    mycomp, mycompname = self.modeler.components.create_inductor(name, value, xpos, ypos,
                                                                                 use_instance_id_netlist=use_instance)
                elif fields[0][0] == "C":
                    if "{" in fields[3][0]:
                        value = fields[3].strip()[1:-1]
                    elif '/' in fields[3] and '"' not in fields[3][0] and "'" not in fields[3][0]:
                        value = self._get_number_from_string(fields[3].split('/')[0])
                    else:
                        value = self._get_number_from_string(fields[3])
                    mycomp, mycompname = self.modeler.components.create_capacitor(name, value, xpos, ypos,
                                                                                  use_instance_id_netlist=use_instance)
                elif fields[0][0] == "Q" or fields[0][0] == "U":
                    if len(fields) == 4 and fields[0][0] == "Q":
                        value = fields[3].strip()
                        mycomp, mycompname = self.modeler.components.create_npn(fields[0], value, xpos, ypos,
                                                                                use_instance_id_netlist=use_instance)
                    else:
                        numpins = len(fields) - 2
                        i = 1
                        pins = []
                        while i <= numpins:
                            pins.append("Pin" + str(i))
                            i += 1
                        parameter = fields[len(fields) - 1][:-1].strip()
                        if "=" in parameter:
                            parameter_list = [parameter[:parameter.find("=") - 1]]
                            parameter_value = [parameter[parameter.find("=") + 1:]]
                        else:
                            parameter_list = ["MOD"]
                            parameter_value = [parameter]
                        self.modeler.components.create_symbol(parameter, pins)
                        already_exist = False
                        for el in self.modeler.components.components:
                            if self.modeler.components.components[el].name == parameter:
                                already_exist = True
                        if not already_exist:
                            self.modeler.components.create_new_component_from_symbol(parameter, pins, fields[0][0],
                                                                                     parameter_list, parameter_value)
                        mycomp, mycompname = self.modeler.components.create_component(fields[0], component_library=None,
                                                                                      component_name=parameter,
                                                                                      xpos=xpos, ypos=ypos,
                                                                                      use_instance_id_netlist=use_instance)
                        value = None
                elif fields[0][0] == "J":
                    numpins = len(fields) - 1
                    i = 1
                    pins = []
                    while i <= numpins:
                        pins.append("Pin" + str(i))
                        i += 1
                    parameter = fields[len(fields) - 1][:-1].strip()
                    if "=" in parameter:
                        parameter_list = [parameter[:parameter.find("=") - 1]]
                        parameter_value = [parameter[parameter.find("=") + 1:]]
                    else:
                        parameter_list = ["MOD"]
                        parameter_value = [parameter]
                    self.modeler.components.create_symbol(parameter, pins)
                    already_exist = False
                    for el in self.modeler.components.components:
                        if self.modeler.components.components[el].name == parameter:
                            already_exist = True
                    if not already_exist:
                        self.modeler.components.create_new_component_from_symbol(parameter, pins, fields[0][0],
                                                                                 parameter_list, parameter_value)
                    mycomp, mycompname = self.modeler.components.create_component(fields[0], component_library=None,
                                                                                  component_name=parameter,
                                                                                  xpos=xpos, ypos=ypos,
                                                                                  use_instance_id_netlist=use_instance)
                    value = None
                elif fields[0][0] == "D":
                    value = self._get_number_from_string(fields[3])
                    mycomp, mycompname = self.modeler.components.create_diode(name, value, xpos, ypos,
                                                                              use_instance_id_netlist=use_instance)
                elif fields[0][0] == "V":
                    if "PULSE" not in line:
                        value = self._get_number_from_string(fields[3])
                        mycomp, mycompname = self.modeler.components.create_voltage_dc(name, value, xpos, ypos,
                                                                                       use_instance_id_netlist=use_instance)
                    else:
                        value = line[line.index("PULSE") + 6:line.index(")") - 1].split(" ")
                        value = [i.replace("{", "").replace("}", "") for i in value]
                        fields[1], fields[2] = fields[2], fields[1]
                        mycomp, mycompname = self.modeler.components.create_voltage_pulse(name, value, xpos, ypos,
                                                                                          use_instance_id_netlist=use_instance)
                elif fields[0][0] == "K":
                    value = self._get_number_from_string(fields[3])
                    mycomp, mycompname = self.modeler.components.create_coupling_inductors(name, fields[1], fields[2],
                                                                                           value, xpos, ypos,
                                                                                           use_instance_id_netlist=use_instance)
                elif fields[0][0] == "I":
                    if "PULSE" not in line:
                        value = self._get_number_from_string(fields[3])
                        mycomp, mycompname = self.modeler.components.create_current_dc(name, value, xpos, ypos,
                                                                                       use_instance_id_netlist=use_instance)
                    else:
                        value = line[line.index("PULSE") + 6:line.index(")") - 1].split(" ")
                        value = [i.replace("{", "").replace("}", "") for i in value]
                        mycomp, mycompname = self.modeler.components.create_current_pulse(name, value, xpos, ypos,
                                                                                          use_instance_id_netlist=use_instance)
                if mycomp:
                    pins = self.modeler.components.get_pins(mycomp)
                    id = 1
                    for pin in pins:
                        pos = self.modeler.components.get_pin_location(mycomp, pin)
                        if pos[0] < xpos:
                            angle = 6.28318530717959
                        else:
                            angle = 3.14159265358979
                        self.modeler.components.create_page_port(fields[id], pos[0], pos[1], angle)
                        id += 1
                    ypos += delta
                    if ypos > 0.254:
                        xpos += delta
                        ypos = 0
                    counter += 1
                    if counter > 59:
                        self.modeler.oeditor.CreatePage("<Page Title>")
                        counter = 0
        if autosave:
            self._desktop.EnableAutoSave(True)
        return True

    @aedt_exception_handler
    def create_schematic_from_mentor_netlist(self, file_to_import):
        """Create a circuit schematic from a Mentor net list.
        
        Supported currently:
        
        * R
        * L
        * C
        * Diodes
        * Bjts
        * Discrete components with syntax ``Uxxx net1 net2 ... netn modname``

        Parameters
        ----------
        file_to_import : str
            Full path to the HSpice file to import.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        xpos = 0
        ypos = 0
        delta = 0.0508
        use_instance = True
        my_netlist = []
        with open(file_to_import, 'r') as f:
            for line in f:
                my_netlist.append(line.split(" "))
        nets = [i for i in my_netlist if i[0] == "NET"]
        comps = [i for i in my_netlist if i[0] == "COMP:"]
        props = {}
        for el in my_netlist:
            if el[0] == "COMP:":
                n = el[2].strip()
                n = n[1:-1]
                props[n] = []
                i = my_netlist.index(el) + 1
                finished = False
                while not finished and i < len(my_netlist):
                    if my_netlist[i][0] == "Property:":
                        props[n].append(my_netlist[i][1])
                    elif "Pin:" in my_netlist[i]:
                        props[n].append(my_netlist[i])
                    else:
                        finished = True
                    i += 1

        column_number = int(math.sqrt(len(comps)))
        for el in comps:
            name = el[2].strip()  # Remove carriage return.
            name = name[1:-1]  # Remove quotes.
            if len(el) > 3:
                comptype = el[3]
            else:
                comptype = self.retrieve_mentor_comp(el[2])
            value = "required"
            for prop in props[name]:
                if "Value=" in prop:
                    value = prop.split("=")[1].replace(",", ".").strip()

            mycomp = None
            if "resistor:RES." in comptype:
                mycomp, mycompname = self.modeler.components.create_resistor(name, value, xpos, ypos,
                                                                             use_instance_id_netlist=use_instance)
            elif "inductor:COIL." in comptype:
                mycomp, mycompname = self.modeler.components.create_inductor(name, value, xpos, ypos,
                                                                             use_instance_id_netlist=use_instance)
            elif "capacitor:CAP." in comptype:
                mycomp, mycompname = self.modeler.components.create_capacitor(name, value, xpos, ypos,
                                                                              use_instance_id_netlist=use_instance)
            elif "transistor:NPN" in comptype:
                mycomp, mycompname = self.modeler.components.create_npn(name, value, xpos, ypos,
                                                                        use_instance_id_netlist=use_instance)
            elif "transistor:PNP" in comptype:
                mycomp, mycompname = self.modeler.components.create_pnp(name, value, xpos, ypos,
                                                                        use_instance_id_netlist=use_instance)
            elif "diode:" in comptype:
                mycomp, mycompname = self.modeler.components.create_diode(name, value, xpos, ypos,
                                                                          use_instance_id_netlist=use_instance)

            if mycomp:
                pins = self.modeler.components.get_pins(mycomp)
                id = 1
                for pin in pins:
                    pos = self.modeler.components.get_pin_location(mycomp, pin)
                    if pos[0] < xpos:
                        angle = 6.28318530717959
                    else:
                        angle = 3.14159265358979
                    netname = None
                    for net in nets:
                        net = [i.strip() for i in net]
                        if (name + "-" + str(id)) in net:
                            fullnetname = net[2]
                            netnames = fullnetname.split("/")
                            netname = netnames[len(netnames) - 1].replace(",", "_").replace("'", "").replace("$",
                                                                                                             "").strip()
                    if not netname:
                        prop = props[name]
                        if "Pin:" in prop and id in prop:
                            netname = prop[-1]
                            netname = netname.replace("$", "")

                    if netname:
                        self.modeler.components.create_page_port(netname, pos[0], pos[1], angle)
                    else:
                        self._messenger.add_info_message("Page Port Not Created", "Global")
                    id += 1
                ypos += delta
                if ypos > delta * (column_number):
                    xpos += delta
                    ypos = 0

        for el in nets:
            netname = el[2][1:-1]
            netname = netname.replace("$", "")
            if "GND" in netname.upper():
                self.modeler.components.create_gnd(xpos, ypos)
                page_pos = ypos + 0.00254
                id, name = self.modeler.components.create_page_port(netname, xpos, ypos, 6.28318530717959)
                mod1 = self.modeler.components[id]
                mod1.set_location(str(xpos) + "meter", str(page_pos) + "meter")
                ypos += delta
                if ypos > delta * column_number:
                    xpos += delta
                    ypos = 0

        return True

    @aedt_exception_handler
    def retrieve_mentor_comp(self, refid):
        """Retrieve the type of the Mentor net list component for a given reference ID.

        Parameters
        ----------
        refid : str
            Reference ID.

        Returns
        -------
        str
            Type of Mentor net list comoponet.
        
        """
        if refid[1] == "R":
            return "resistor:RES."
        elif refid[1] == "C":
            return "capacitor:CAP."
        elif refid[1] == "L":
            return "inductor:COIL."
        elif refid[1] == "D":
            return "diode"
        elif refid[1] == "Q":
            return "transistor:NPN"
        else:
            return ""

    @aedt_exception_handler
    def get_source_pin_names(self, source_design_name, source_project_name=None, source_project_path=None,
                             port_selector=3):
        """List the pin names.
        
        Parameters
        ----------
        source_design_name : str
            Name of the source design.
        source_project_name :str, optional
            Name of the source project. The default is ``None``.
        source_project_path : str, optional
            Path to the source project if different than the existing path. The default is ``None``.
        port_selector : int, optional
             Type of the port. Options are ``1``, ``2``, or ``3``, corresponding respectively to ``"Wave Port"``, ``"Terminal"``, or ``"Circuit Port"``.
             The default is ``3``, which is a circuit port.

        Returns
        -------
        list
            List of pin names.
        
        """
        if not source_project_name or self.project_name == source_project_name:
            oSrcProject = self._desktop.GetActiveProject()
        else:
            self._desktop.OpenProject(source_project_path)
            oSrcProject = self._desktop.SetActiveProject(source_project_name)
        oDesign = oSrcProject.SetActiveDesign(source_design_name)
        oModule = oDesign.GetModule("BoundarySetup")
        port = None
        if port_selector == 1:
            port = "Wave Port"
        elif port_selector == 2:
            port = "Terminal"
        elif port_selector == 3:
            port = "Circuit Port"
        if not port:
            return False
        pins = list(oModule.GetExcitationsOfType(port))

        return pins

    @aedt_exception_handler
    def import_touchstone_solution(self, filename, solution_name="Imported_Data"):
        """Import a Touchstone file as the solution.

        Parameters
        ----------
        filename : str
            Name of the Touchstone file.
        solution_name : str, optional
            Name of the solution. The default is ``"Imported_Data"``.
        
        Returns
        -------
        list
            List of ports.
        
        """
        if filename[-2:] == "ts":
            with open(filename, "r") as f:
                lines = f.readlines()
                for i in lines:
                    if "[Number of Ports]" in i:
                        ports = int(i[i.find("]") + 1:])
                portnames = [i.split(" = ")[1].strip() for i in lines if "! Port" in i[:9]]
                if not portnames:
                    portnames = ["Port{}".format(i + 1) for i in range(ports)]
        else:
            re_filename = re.compile(r"\.s(?P<ports>\d+)+p", re.I)
            m = re_filename.search(filename)
            ports = int(m.group('ports'))
            portnames = None
            with open(filename, "r") as f:
                lines = f.readlines()
                portnames = [i.split(" = ")[1].strip() for i in lines if "Port[" in i]
            if not portnames:
                portnames = ["Port{}".format(i + 1) for i in range(ports)]
        arg = ["NAME:NPortData", "Description:=", "", "ImageFile:=", "",
               "SymbolPinConfiguration:=", 0, ["NAME:PortInfoBlk"], ["NAME:PortOrderBlk"],
               "filename:=", filename, "numberofports:=", ports, "sssfilename:=", "",
               "sssmodel:=", False, "PortNames:=", portnames,
               "domain:=", "frequency", "datamode:=", "Link", "devicename:=", "",
               "SolutionName:=", solution_name, "displayformat:=", "MagnitudePhase", "datatype:=", "SMatrix",
               ["NAME:DesignerCustomization",
                "DCOption:=", 0, "InterpOption:=", 0, "ExtrapOption:=", 1,
                "Convolution:=", 0, "Passivity:=", 0, "Reciprocal:=", False,
                "ModelOption:=", "", "DataType:=", 1],
               ["NAME:NexximCustomization", "DCOption:=", 3, "InterpOption:=", 1,
                "ExtrapOption:=", 3, "Convolution:=", 0, "Passivity:=", 0,
                "Reciprocal:=", False, "ModelOption:=", "", "DataType:=", 2],
               ["NAME:HSpiceCustomization", "DCOption:=", 1, "InterpOption:=", 2,
                "ExtrapOption:=", 3, "Convolution:=", 0, "Passivity:=", 0,
                "Reciprocal:=", False, "ModelOption:=", "", "DataType:=", 3],
               "NoiseModelOption:=", "External"]
        self.odesign.ImportData(arg, "", True)
        return portnames

    @aedt_exception_handler
    def export_touchstone(self, solutionname, sweepname, filename=None, variation=[], variations_value=[]):
        """Export the Touchstone file to a local folder.

        Parameters
        ----------
        solutionname : str
             Name of the solution that has been solved.
        sweepname : str
             Name of the sweep that has been solved.
        filename : str, optional
             Full path to the output file. The default is ``None``.
        variation : list, optional
             List of all parameter variations. For example, ``["$AmbientTemp", "$PowerIn"]``.
             The default is ``[]``.
        variations_value : list, optional
             List of all parameter variation values. For example, ``["22cel", "100"]``.
             The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        # Normalize the save path
        if not filename:
            appendix = ""
            for v, vv in zip(variation, variations_value):
                appendix += "_" + v + vv.replace("\'", "")
            ext = ".S" + str(self.oboundary.GetNumExcitations()) + "p"
            filename = os.path.join(self.project_path, solutionname + "_" + sweepname + appendix + ext)
        else:
            filename = filename.replace("//", "/").replace("\\", "/")
        print("Exporting Touchstone " + filename)
        DesignVariations = ""
        i = 0
        for el in variation:
            DesignVariations += str(variation[i]) + "=\'" + str(variations_value[i].replace("\'", "")) + "\' "
            i += 1
            # DesignVariations = "$AmbientTemp=\'22cel\' $PowerIn=\'100\'"
        # array containing "SetupName:SolutionName" pairs (note that setup and solution are separated by a colon)
        SolutionSelectionArray = [solutionname + ":" + sweepname]
        # 2=tab delimited spreadsheet (.tab), 3= touchstone (.sNp), 4= CitiFile (.cit),
        # 7=Matlab (.m), 8=Terminal Z0 spreadsheet
        FileFormat = 3
        OutFile = filename  # full path of output file
        FreqsArray = ["all"]  # array containin the frequencies to export, use ["all"] for all frequencies
        DoRenorm = True  # perform renormalization before export
        RenormImped = 50  # Real impedance value in ohm, for renormalization
        DataType = "S"  # Type: "S", "Y", or "Z" matrix to export
        Pass = -1  # The pass to export. -1 = export all passes.
        ComplexFormat = 0  # 0=Magnitude/Phase, 1=Real/Immaginary, 2=dB/Phase
        DigitsPrecision = 15  # Touchstone number of digits precision
        IncludeGammaImpedance = True  # Include Gamma and Impedance in comments
        NonStandardExtensions = False  # Support for non-standard Touchstone extensions

        self.odesign.ExportNetworkData(DesignVariations, SolutionSelectionArray, FileFormat,
                                         OutFile, FreqsArray, DoRenorm, RenormImped, DataType, Pass,
                                         ComplexFormat, DigitsPrecision, False, IncludeGammaImpedance,
                                         NonStandardExtensions)
        return True

    @aedt_exception_handler
    def export_fullwave_spice(self, designname=None, setupname=None, is_solution_file=False, filename=None,
                              passivity=False, causality=False, renormalize=False, impedance=50, error=0.5,
                              poles=10000):
        """
        Export a full wave HSpice file using NDE.

        .. warning::
          This method doesn't work.

        Parameters
        ----------
        designname : str, optional
            Name of the design or the full path to the solution file if it is an imported file.
            The default is ``None``.    
        setupname : str, optional
            Name of the setup if it is a design. The default is ``None``.
        is_solution_file: bool, optional
            Whether it is an imported solution file. The default is ``False``. 
        filename: str, optional
            Full path and name for exporting the HSpice file. The default is ``None``.
        passivity: bool, optional
            Whether to compute the passivity. The default is ``False``.
        causality: bool, optional
            Whether to compute the causality. The default is ``False``.
        renormalize: bool, optional
            Whether to renormalize the S-matrix to a specific port impedance.
            The default is ``False``.
        impedance: float, optional
            Impedance value if ``renormalize=True``. The default is ``50``.
        error: float, optional
            Fitting error. The default is ``0.05``.
        poles: int, optional
            Number of fitting poles. The default is ``10000``.

        Returns
        -------
        str
            File name if the export is successful.
        
        """
        if not designname:
            designname = self.design_name
        if not filename:
            filename = os.path.join(self.project_path, self.design_name + ".sp")
        if is_solution_file:
            setupname = designname
            designname = ""
        else:
            if not setupname:
                setupname = self.nominal_sweep
        self.onetwork_data_explorer.ExportFullWaveSpice(designname, is_solution_file, setupname, "",
                                                        [],
                                                        ["NAME:SpiceData", "SpiceType:=", "HSpice",
                                                         "EnforcePassivity:=", passivity, "EnforceCausality:=",
                                                         causality,
                                                         "UseCommonGround:=", True,
                                                         "ShowGammaComments:=", True,
                                                         "Renormalize:=", renormalize,
                                                         "RenormImpedance:=", impedance,
                                                         "FittingError:=", error,
                                                         "MaxPoles:=", poles,
                                                         "PassivityType:=", "IteratedFittingOfPV",
                                                         "ColumnFittingType:=", "Matrix",
                                                         "SSFittingType:=", "FastFit",
                                                         "RelativeErrorToleranc:=", False,
                                                         "EnsureAccurateZfit:=", True,
                                                         "TouchstoneFormat:=", "MA",
                                                         "TouchstoneUnits:=", "GHz",
                                                         "TouchStonePrecision:=", 15,
                                                         "SubcircuitName:=", "",
                                                         "SYZDataInAutoMode:=", False,
                                                         "ExportDirectory:=", os.path.dirname(filename) + "\\",
                                                         "ExportSpiceFileName:=", os.path.basename(filename),
                                                         "FullwaveSpiceFileName:=",
                                                         os.path.basename(filename), "UseMultipleCores:=",
                                                         True, "NumberOfCores:=", 20])
        return filename

    @aedt_exception_handler
    def create_touchstone_report(self, plot_name, curvenames, solution_name=None, variation_dict=None):
        """Create a Touchstone plot.

        Parameters
        ----------
        plot_name : str
            Name of the plot.
        curvenames : str
            List of the curves to plot.
        solution_name : str, optional
            Name of the solution. The default value is ``None``.
        variation_dict : dict, optional
            Dictionary of variation names. The default value is ``None``.

        Returns
        -------
        bool
           ``True`` when successful, ``False`` when failed.

        """
        if not solution_name:
            solution_name = self.nominal_sweep
        variations = ["Freq:=", ["All"]]
        if variation_dict:
            for el in variation_dict:
                variations.append(el + ":=")
                variations.append([variation_dict[el]])
        self.post.oreportsetup.CreateReport(plot_name, "Standard", "Rectangular Plot", solution_name,
                                            ["NAME:Context", "SimValueContext:=",
                                             [3, 0, 2, 0, False, False, -1, 1, 0, 1, 1, "", 0, 0]], variations,
                                            ["X Component:=", "Freq", "Y Component:=", curvenames])
        return True



from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from .Analysis import Analysis
from .Design import solutions_settings, Design
from ..modeler.Circuit import ModelerEmit
from ..modules.PostProcessor import PostProcessor
from ..modules.SetupTemplates import SetupKeys
from ..modules.SolveSetup import SetupCircuit


class FieldAnalysisEmit(Design):
    """FieldAnaysisEmit class.

    The class is for setting up an EMIT analysis in AEDT.
    It is automatically initialized by an application call (like for HFSS,
    Q3D, and other tools). Refer to the Application function for input definitions.

    
    """

    @property
    def solution_type(self):
        """Solution type.
          
        Parameters
        ----------
        soltype :
            SolutionType object.
        """
        return self._solution_type

    @solution_type.setter
    def solution_type(self, soltype):
        """ """

        if soltype:
            self._solution_type = "EMIT"
        else:
            self._solution_type = "EMIT"

    @property
    def existing_analysis_setups(self):
        """Existing analysis setups."""
        return []

    @property
    def setup_names(self):
        """Setup names."""
        return []


    def __init__(self, application, projectname, designname, solution_type, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=True, release_on_exit=True, student_version=False):
        self.solution_type = solution_type
        Design.__init__(self, application, projectname, designname, solution_type,
                        specified_version, NG, AlwaysNew, release_on_exit, student_version)
        self._modeler = ModelerEmit(self)
        self._post = PostProcessor(self)

    @property
    def modeler(self):
        """Modeler.
        
        Returns
        -------
        :class:`pyaedt.modeler.Circuit.ModelerEmit`
            Design oModeler
        """
        return self._modeler

    @property
    def oanalysis(self):
        """Analysis object.
        
        Returns
        -------
        int
           Design module ``"SimSetup"``
        """
        return None
